function DispGradientLines(im, param)
% Print the gradient lines on an image :
% size(im_grad) = size(im)
% Graphical display settings are encoded in param
%
% param :
%       param.display.markertype : string
%           See scatter doc for list of values, such as 'none', '+', ...
%       param.display.line : 0 or 1
%           Should we plot a line between the points ?
%       param.display.line_alpha : 0 to 1
%           transparency of the line
%       param.display.linespec : string
%           Line parameters : see plot doc.
%       param.display.linewidth : int
%           Line parameters : see plot doc.
%       param.display.quiver : 0 or 1
%           Should we plot small arrows ?
%       param.display.scatter : 0 or 1
%           Should we plot points ?
%       param.display.scatter_alpha : 0 to 1
%           transparency of the scatter
%       param.display.area_param = {'const', 'grad', 'arc-length', 'time'}
%           Which parameters controls the area of the markers
%       param.display.area_fun : param -> area function handle
%           The function that maps the area_param to an area, in pix^2
%       param.display.color_param = {'const', 'grad', 'arc-length', 'time'}
%           Which parameters controls the color of the markers
%       param.display.color_fun : param -> [R;G;B] function handle
%           The function that maps the color_param to a RGB color
%
% Author : Jean Feydy

im = double(im); % VERY IMPORTANT, otherwise, negative values won't be allowed for the gradient...
[lines, grad] = GetGradientLines(im, param);

% Plotting
hold on;
set(gca,'YDir','reverse');
DispBackgroundImage(im, param);
fprintf('Background image displayed\n');
if param.display.quiver
    hold on;
    quiver(grad(:,:,1), grad(:,:,2), 4);
    hold off;
end;
for i = 1:numel(lines)
    fprintf('l ');
    if isequal(mod(i, 50), 0)
        fprintf('\n');
    end;
    line = lines{i};
    line = line + 0.5; % Matlab centers the pixels of the plotted images
    if param.display.line
        if param.display.line_alpha >= 1
            plot(line(1,:), line(2,:), param.display.linespec, 'LineWidth', param.display.linewidth);
        else
            xflip = [line(1, 1 : end - 1) fliplr(line(1,:))];
            yflip = [line(2, 1 : end - 1) fliplr(line(2,:))];
            patch(xflip, yflip, 'b', 'EdgeColor','b', 'EdgeAlpha', param.display.line_alpha , 'LineWidth', param.display.linewidth, 'FaceColor', 'none');
        end;
    end;
    if param.display.scatter
        area_param = GetParameterAlongLine(param.display.area_param, line);
        area_param = cellfun(param.display.area_fun, num2cell(area_param, 1), 'UniformOutput', false);
        area_param = cell2mat(area_param);
        
        color_param = GetParameterAlongLine(param.display.color_param, line);
        color_param = cellfun(param.display.color_fun, num2cell(color_param, 1), 'UniformOutput', false);
        color_param = cell2mat(color_param)';
        
        scatter(line(1,:), line(2,:), area_param, color_param, 'Marker', param.display.markertype);
    end;
end;
fprintf('\n')
axis equal;
hold off;


    function p = GetParameterAlongLine(s, l)
        % s might be {'const', 'grad', 'arc-length', 'time'},
        % l is the list of [x;y] points.
        n_points = numel(l(1,:));
        if strcmp(s, 'const')
            p = ones(1, n_points);
        elseif strcmp(s, 'grad')
            %grad = ComputeGradient(im, param);
            %grad = ProcessGradient(grad, param);
            f = @Anonymous_GetGrad;
            p = cellfun(f, num2cell(l, 1), 'UniformOutput', false);
            p = cell2mat(p);
        elseif strcmp(s, 'arc-length')
            di = l(:, 2:end) - l(:, 1:end-1);
            d = sqrt(di(1,:).^2 + di(2,:).^2);
            p = [0, cumsum(d)];
        end;
        
        function g = Anonymous_GetGrad(x)
            g = GetGradientFromFloatPoint(x, grad, param);
        end
    end

end




