function [next_point, gr, dt] = GetNextPoint(prev_point, grad, param)
    % Function where cauchy/runge-kutta & co. are implemented.
    %
    %       param.computeline.method = {'cauchy', 'explicit-bilinear',
    %           'explicit-piecewise', 'runge-kutta-4'}
    %       param.computeline.time_step : double
    %           N.B. : typically, around 0.1
    %       param.computeline.stopping.time.type = {'time', 'arc-length'}
    %           N.B. : this is the exploration time in each direction
    %                  if 'arc-length', then 'dt' = ds
    %
    % Author : Jean Feydy
    
    [SY, SX, ~] = size(grad);
    
    if strcmp(param.computeline.method, 'cauchy')
        next_point = CauchyStep(prev_point);
        gr = GetGradientFromFloatPoint(next_point, grad, param);
        %n_i = ToGrid(next_point);
        %gr = grad(n_i(2), n_i(1), :);
        %gr = permute(gr, [3,1,2]);
    elseif strcmp(param.computeline.method, 'explicit-piecewise')
        next_point = ExplicitPiecewiseStep(prev_point);
        n_i = ToGrid(next_point);
        gr = grad(n_i(2), n_i(1), :);
        gr = permute(gr, [3,1,2]);
    end;
    
    if strcmp(param.computeline.stopping.time.type, 'time')
        dt = param.computeline.time_step;
    elseif strcmp(param.computeline.stopping.time.type, 'arc-length')
        dt = norm(prev_point - next_point);
    else
        error('param.computeline.stopping.time.type must be either time or arc-length');
    end;
    
    function x_i = ToGrid(x_f)
        % Maps a float point to the grid, in order to simply get values
        % such as grad.
        
        x_i = max(1, ceil(x_f));
        x_i = min(x_i, [SX; SY]);
    end
    
    function x_b = StayInFrame(x)
        x_b = max(0, min(x, [SX;SY]));
    end
    
    function n = CauchyStep(p)
        % Executes a simple Cauchy step
        
        %p_i = ToGrid(p);
        %g = permute(grad(p_i(2), p_i(1), :), [3,1,2]);
        g = GetGradientFromFloatPoint(p, grad, param);
        n = StayInFrame(p + param.computeline.time_step * g);
    end
    function n = ExplicitPiecewiseStep(pr)
        % Executes an explicite piecewise step :
        % The gradient is supposed to be piecewise constant in every pixel:
        % for p in ]x-1, x] x ]y-1, y], where x, y integers,
        % the corresponding discrete pixel is (x,y), and the gradient is
        % equal to grad(ToGrid(p)) (with the usual 3-lines conversion).
        %
        % It is therefore possible to explicitely compute the coordinates
        % of the point on the border of the pixel where we end up.
        % The resulting trajectory will be piecewise linear, zig-zagging
        % between the pixel borders.
        %
        % Author : Jean Feydy
        
        
        EPS = 0.001;
        t_m = 0;
        p = pr;
        n = [0;0];
        if abs(p - round(p)) <= EPS/2
            p = round(p);
            CornerStep()
        elseif abs(p(1) - round(p(1))) <= EPS/2
            p(1) = round(p(1));
            VertSideStep()
        elseif abs(p(2) - round(p(2))) <= EPS/2
            p(2) = round(p(2));
            HorzSideStep()
        else
            InsideStep()
        end;
        
        function g = GetPGr (x)
            x_i = ToGrid(x);
            g = grad(x_i(2), x_i(1), :);
            g = permute(g, [3,1,2]);            
        end
        
        function CornerStep()
            x = p(1);
            y = p(2);
            g_tl = GetPGr([x;y]);
            g_tr = GetPGr([x+1;y]);
            g_bl = GetPGr([x;y+1]);
            g_br = GetPGr([x+1;y+1]);
            m = g_tl + g_tr + g_bl + g_br;
            p = p + 2*EPS*m;
            InsideStep();
        end
        
        function VertSideStep()
            x = round(p(1)); % ~= p(1)
            y = p(2);
            g_l = GetPGr([x;y]);
            g_r = GetPGr([x+1;y]);
            if g_l(1) >= 0 && g_r(1) > 0
                p = p + [EPS; 0];
                AdvanceAlong(g_r);
            elseif g_l(1) < 0 && g_r(1) <= 0
                p = p + [-EPS; 0];
                AdvanceAlong(g_l)
            elseif g_l(1) >= 0 && g_r(1) <= 0
                % Both vectors force to stay on the Vertical Edge
                n_g = [0; -g_r(1)*g_l(2) + g_l(1)*g_r(2) + EPS];
                p = p + EPS * sign(n_g);
                AdvanceAlong(n_g) % That choice may be discussed
            elseif g_l(1) < 0  && g_r(1) > 0
                % We must leave the edge, but in which direction ?
                if g_l(1) + g_r(1) >= 0 % let's say g_r wins
                    p = p + [EPS; 0];
                    AdvanceAlong(g_r);
                else
                    p = p + [-EPS; 0];
                    AdvanceAlong(g_l);
                end;
            else
                error('Well, this was not supposed to happen...');
            end;
        end
        
        function HorzSideStep()
            x = p(1);
            y = round(p(2)); % ~= p(2)
            g_b = GetPGr([x;y+1]);
            g_t = GetPGr([x;y]);
            if g_b(2) <= 0 && g_t(2) < 0
                p = p + [0;-EPS];
                AdvanceAlong(g_t);
            elseif g_b(2) > 0 && g_t(2) >= 0
                p = p + [0;EPS];
                AdvanceAlong(g_b)
            elseif g_b(2) <= 0 && g_t(2) >= 0
                % Both vectors force to stay on the Horizontal Edge
                n_g = [g_t(2)*g_b(1) - g_b(2)*g_t(1) + EPS; 0];
                p = p + EPS * sign(n_g);
                AdvanceAlong(n_g) % That choice may be discussed
            elseif g_b(2) > 0  && g_t(2) < 0
                % We must leave the edge, but in which direction ?
                if g_b(2) + g_t(2) >= 0 % let's say g_t wins
                    p = p + [0;-EPS];
                    AdvanceAlong(g_t);
                else
                    p = p + [0;EPS];
                    AdvanceAlong(g_b);
                end;
            else
                error('Well, this was not supposed to happen...');
            end;
        end
        
        function InsideStep()
            g = GetPGr(p);
            AdvanceAlong(g);
        end
        
        function AdvanceAlong(g)
            p_i = ToGrid(p);
            t = [0;0];
            if isequal(g, [0;0])
                t_m = 0;
                n = p;
            elseif isequal(g(1), 0)
                t_m = (p_i(2) - p(2) - (g(2) < 0) ) ./ g(2);
                n = p + t_m * g;
            elseif isequal(g(2), 0)
                t_m = (p_i(1) - p(1) - (g(1) < 0) ) ./ g(1);
                n = p + t_m * g;
            else
                t = (p_i - p - (g < 0) ) ./ g;
                % t contains the times before reaching the sides
                t_m = min(t(1), t(2));
                n = p + t_m * g;
            end;
        end
        
    end
%     function n = ExplicitPiecewiseStep(p)
%         % Executes an explicite piecewise step :
%         % The gradient is supposed to be piecewise constant in every pixel:
%         % for p in ]x-1, x] x ]y-1, y], where x, y integers,
%         % the corresponding discrete pixel is (x,y), and the gradient is
%         % equal to grad(ToGrid(p)) (with the usual 3-lines conversion).
%         %
%         % It is therefore possible to explicitely compute the coordinates
%         % of the point on the border of the pixel where we end up.
%         % The resulting trajectory will be piecewise linear, zig-zagging
%         % between the pixel borders.
%         %
%         % Author : Jean Feydy
%         
%         p_i = ToGrid(p);
%         gr_p = grad(p_i(2), p_i(1), :);
%         gr_p = permute(gr_p, [3,1,2]);
%         
%         t = [0;0];
%         if isequal(gr_p, [0;0])
%             t_m = 0;
%             n = p;
%         elseif isequal(gr_p(1), 0)
%             t_m = (p_i(2) - p(2) - (gr_p(2) < 0) ) ./ gr_p(2);
%             n = p + t_m * gr_p;
%         elseif isequal(gr_p(2), 0)
%             t_m = (p_i(1) - p(1) - (gr_p(1) < 0) ) ./ gr_p(1);
%             n = p + t_m * gr_p;
%         else
%             t = (p_i - p - (gr_p < 0) ) ./ gr_p;
%             % t contains the times before reaching the sides
%             if isequal( abs(p_i - p - (gr_p < 0) ) > 0.002 , [0;0])
%                 n = p;
%                 % We are in a corner, and will go to the opposite one
%                 % We have to check if we will come back immediately,
%                 % In which case we launch a stopping exception
%                 n_t = StayInFrame(n + 0.0001 * sign(gr_p));
%                 n_t_i = ToGrid(n_t);
%                 gr_n_t = grad(n_t_i(2), n_t_i(1), :);
%                 gr_n_t = permute(gr_n_t, [3,1,2]);
%                 if gr_n_t' * gr_p < 0
%                     throw(MException('GetNextPoint:StopLine', 'ExplicitPiecewiseStep : Critical point reached'));
%                 end;
%                 
%             elseif isequal(abs(t(1) * gr_p(1)) > 0.002 , 0)
%                 t_m = t(2);
%                 n = p + t_m * [0;gr_p(2)];
%             elseif isequal(abs(t(2) * gr_p(2)) > 0.002, 0)
%                 t_m = t(1);
%                 n = p + t_m * [gr_p(1); 0];
%             else
%                 t_m = min(t(1), t(2));
%                 n = p + t_m * gr_p;
%             end;
%         end;
%         
%         n = StayInFrame(n + 0.0001 * sign(gr_p));
%     end
 end




